##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################
import pprint
import uuid

import numpy as np
from zlmdb import MapBytes20TimestampUuid, MapUuidBytes20Bytes20Uint8UuidFlatBuffers, flatbuffers, table

from cfxdb import pack_uint256, unpack_uint256
from cfxdb.gen.xbr import Consent as ConsentGen


class _ConsentGen(ConsentGen.Consent):
    @classmethod
    def GetRootAsConsent(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _ConsentGen()
        x.Init(buf, n + offset)
        return x

    def _as_bytes(self, offet):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(offet))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def MarketOidAsBytes(self):
        return self._as_bytes(4)

    def MemberAsBytes(self):
        return self._as_bytes(6)

    def DelegateAsBytes(self):
        return self._as_bytes(8)

    def CatalogOidAsBytes(self):
        return self._as_bytes(12)

    def UpdatedAsBytes(self):
        return self._as_bytes(16)

    def TidAsBytes(self):
        return self._as_bytes(22)

    def SignatureAsBytes(self):
        return self._as_bytes(24)


class Consent(object):
    """
    ``XBRNetwork.Consent`` database object.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # [uint8] (uuid);
        self._market_oid = None

        # [uint8] (address)
        self._member = None

        # [uint8] (address)
        self._delegate = None

        # [uint8] ActorType
        self._delegate_type = None

        # [uint8] (uuid);
        self._catalog_oid = None

        # uint64 (timestamp)
        self._timestamp = None

        # [uint8] (uint256)
        self._updated = None

        # bool
        self._consent = None

        # string
        self._service_prefix = None

        # [uint8] (ethhash)
        self._tid = None

        # [uint8] (ethsig)
        self._signature = None

        # bool
        self._synced = None

    def marshal(self) -> dict:
        obj = {
            "market_oid": self.market_oid.bytes if self.market_oid else None,
            "member": bytes(self.member) if self.member else None,
            "delegate": bytes(self.delegate) if self.delegate else None,
            "delegate_type": int(self.delegate_type) if self.delegate_type else None,
            "catalog_oid": self.catalog_oid.bytes if self.catalog_oid else None,
            "timestamp": int(self.timestamp) if self.timestamp else None,
            "updated": pack_uint256(self.updated) if self.updated else None,
            "consent": self.consent,
            "service_prefix": self.service_prefix if self.service_prefix else None,
            "tid": bytes(self.tid) if self.tid else None,
            "signature": bytes(self.signature) if self.signature else None,
            "synced": self.synced,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def market_oid(self) -> uuid.UUID:
        """
        The unique ID of the market.
        """
        if self._market_oid is None and self._from_fbs:
            if self._from_fbs.MarketOidLength():
                _market_oid = self._from_fbs.MarketOidAsBytes()
                self._market_oid = uuid.UUID(bytes=bytes(_market_oid))
        return self._market_oid

    @market_oid.setter
    def market_oid(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._market_oid = value

    @property
    def member(self) -> bytes:
        """
        The unique ID of the market.
        """
        if self._member is None and self._from_fbs:
            if self._from_fbs.MemberLength():
                self._member = self._from_fbs.MemberAsBytes()
        return self._member

    @member.setter
    def member(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 20)
        self._member = value

    @property
    def delegate(self) -> bytes:
        """
        The unique ID of the market.
        """
        if self._delegate is None and self._from_fbs:
            if self._from_fbs.DelegateLength():
                self._delegate = self._from_fbs.DelegateAsBytes()
        return self._delegate

    @delegate.setter
    def delegate(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 20)
        self._delegate = value

    @property
    def delegate_type(self) -> int:
        """
        The unique ID of the market.
        """
        if self._delegate_type is None and self._from_fbs:
            self._delegate_type = self._from_fbs.DelegateType()
        return self._delegate_type

    @delegate_type.setter
    def delegate_type(self, value: int):
        assert value is None or type(value) == int
        self._delegate_type = value

    @property
    def catalog_oid(self) -> uuid.UUID:
        """
        The unique ID of the catalog.
        """
        if self._catalog_oid is None and self._from_fbs:
            if self._from_fbs.CatalogOidLength():
                _catalog_oid = self._from_fbs.CatalogOidAsBytes()
                self._catalog_oid = uuid.UUID(bytes=bytes(_catalog_oid))
        return self._catalog_oid

    @catalog_oid.setter
    def catalog_oid(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._catalog_oid = value

    @property
    def timestamp(self) -> np.datetime64:
        """
        Database transaction time (epoch time in ns) of insert or last update.
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def updated(self) -> int:
        """
        The amount of XBR tokens a XBR provider joining the market must deposit.
        """
        if self._updated is None and self._from_fbs:
            if self._from_fbs.UpdatedLength():
                _updated = self._from_fbs.UpdatedAsBytes()
                self._updated = unpack_uint256(bytes(_updated))
            else:
                self._updated = 0
        return self._updated

    @updated.setter
    def updated(self, value: int):
        assert value is None or type(value) == int
        self._updated = value

    @property
    def consent(self) -> bool:
        """
        Whether we have the consent
        """
        if self._consent is None and self._from_fbs:
            self._consent = self._from_fbs.Consent()
        return self._consent

    @consent.setter
    def consent(self, value: bool):
        assert type(value) == bool
        self._consent = value

    @property
    def service_prefix(self) -> str:
        if self._service_prefix is None and self._from_fbs:
            service_prefix = self._from_fbs.ServicePrefix()
            if service_prefix:
                self.service_prefix = service_prefix.decode("utf8")
        return self._service_prefix

    @service_prefix.setter
    def service_prefix(self, value):
        assert value is None or type(value) == str
        self._service_prefix = value

    @property
    def tid(self) -> bytes:
        """
        Transaction hash of the transaction this change was committed to the blockchain under.
        """
        if self._tid is None and self._from_fbs:
            if self._from_fbs.TidLength():
                self._tid = self._from_fbs.TidAsBytes()
        return self._tid

    @tid.setter
    def tid(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 32)
        self._tid = value

    @property
    def signature(self) -> bytes:
        """
        When signed off-chain and submitted via ``XBRMarket.createMarketFor``.
        """
        if self._signature is None and self._from_fbs:
            if self._from_fbs.SignatureLength():
                self._signature = self._from_fbs.SignatureAsBytes()
        return self._signature

    @signature.setter
    def signature(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 65)
        self._signature = value

    @property
    def synced(self) -> bool:
        """
        Whether the consent is synced with blockchain
        """
        if self._synced is None and self._from_fbs:
            self._synced = self._from_fbs.Synced()
        return self._synced

    @synced.setter
    def synced(self, value: bool):
        assert type(value) == bool
        self._synced = value

    @staticmethod
    def cast(buf):
        return Consent(_ConsentGen.GetRootAsConsent(buf, 0))

    def build(self, builder):
        market_oid = self.market_oid.bytes if self.market_oid else None
        if market_oid:
            market_oid = builder.CreateString(market_oid)

        member = self.member
        if member:
            member = builder.CreateString(member)

        delegate = self.delegate
        if delegate:
            delegate = builder.CreateString(delegate)

        delegate_type = self.delegate_type

        catalog_oid = self.catalog_oid.bytes if self.catalog_oid else None
        if catalog_oid:
            catalog_oid = builder.CreateString(catalog_oid)

        updated = self.updated
        if updated:
            updated = builder.CreateString(pack_uint256(updated))

        service_prefix = self.service_prefix
        if service_prefix:
            service_prefix = builder.CreateString(service_prefix)

        tid = self.tid
        if tid:
            tid = builder.CreateString(tid)

        signature = self.signature
        if signature:
            signature = builder.CreateString(signature)

        ConsentGen.ConsentStart(builder)

        if market_oid:
            ConsentGen.ConsentAddMarketOid(builder, market_oid)

        if member:
            ConsentGen.ConsentAddMember(builder, member)

        if delegate:
            ConsentGen.ConsentAddDelegate(builder, delegate)

        if delegate_type:
            ConsentGen.ConsentAddDelegateType(builder, delegate_type)

        if catalog_oid:
            ConsentGen.ConsentAddCatalogOid(builder, catalog_oid)

        if self.timestamp:
            ConsentGen.ConsentAddTimestamp(builder, int(self.timestamp))

        if updated:
            ConsentGen.ConsentAddUpdated(builder, updated)

        ConsentGen.ConsentAddConsent(builder, self.consent)
        ConsentGen.ConsentAddSynced(builder, self.synced or False)

        if service_prefix:
            ConsentGen.ConsentAddServicePrefix(builder, service_prefix)

        if tid:
            ConsentGen.ConsentAddTid(builder, tid)

        if signature:
            ConsentGen.ConsentAddSignature(builder, signature)

        final = ConsentGen.ConsentEnd(builder)

        return final


@table("8b10071b-5b2a-478f-8101-52dfbaf0760a", build=Consent.build, cast=Consent.cast)
class Consents(MapUuidBytes20Bytes20Uint8UuidFlatBuffers):
    """
    Consents table
    """


@table("7cdac5f2-bddf-4fda-8e3c-21938a0c3667")
class IndexConsentByMemberAddress(MapBytes20TimestampUuid):
    """
    Consent-by-member-address index with ``(member_adr|bytes[20], timestamp|int) -> consent_oid|UUID`` mapping.
    """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# XBR Network members.
class Member(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Member()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsMember(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Member
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Ethereum address of the member.
    # Member
    def Address(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Member
    def AddressAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Member
    def AddressLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Member
    def AddressIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Account OID (on ``planet.xbr.network ``) - only filled when this XBR network member has on-boarded via ``planet.xbr.network`` and has an account there.
    # Member
    def AccountOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Member
    def AccountOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Member
    def AccountOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Member
    def AccountOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Database transaction time (epoch time in ns) of insert or last update.
    # Member
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Block number (on the blockchain) when the member (originally) registered.
    # Member
    def Registered(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Member
    def RegisteredAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Member
    def RegisteredLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Member
    def RegisteredIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # EULA the member agreed to when joining the market (IPFS Multihash string).
    # Member
    def Eula(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Optional member profile (IPFS Multihash string).
    # Member
    def Profile(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Current member level.
    # Member
    def Level(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Transaction hash of the transaction this change was committed to the blockchain under.
    # Member
    def Tid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Member
    def TidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Member
    def TidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Member
    def TidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # When signed off-chain and submitted via ``XBRNetwork.registerMemberFor``.
    # Member
    def Signature(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Member
    def SignatureAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Member
    def SignatureLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Member
    def SignatureIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        return o == 0


def MemberStart(builder):
    builder.StartObject(9)


def Start(builder):
    MemberStart(builder)


def MemberAddAddress(builder, address):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(address), 0)


def AddAddress(builder, address):
    MemberAddAddress(builder, address)


def MemberStartAddressVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAddressVector(builder, numElems: int) -> int:
    return MemberStartAddressVector(builder, numElems)


def MemberAddAccountOid(builder, accountOid):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(accountOid), 0)


def AddAccountOid(builder, accountOid):
    MemberAddAccountOid(builder, accountOid)


def MemberStartAccountOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAccountOidVector(builder, numElems: int) -> int:
    return MemberStartAccountOidVector(builder, numElems)


def MemberAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(2, timestamp, 0)


def AddTimestamp(builder, timestamp):
    MemberAddTimestamp(builder, timestamp)


def MemberAddRegistered(builder, registered):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(registered), 0)


def AddRegistered(builder, registered):
    MemberAddRegistered(builder, registered)


def MemberStartRegisteredVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartRegisteredVector(builder, numElems: int) -> int:
    return MemberStartRegisteredVector(builder, numElems)


def MemberAddEula(builder, eula):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(eula), 0)


def AddEula(builder, eula):
    MemberAddEula(builder, eula)


def MemberAddProfile(builder, profile):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(profile), 0)


def AddProfile(builder, profile):
    MemberAddProfile(builder, profile)


def MemberAddLevel(builder, level):
    builder.PrependUint8Slot(6, level, 0)


def AddLevel(builder, level):
    MemberAddLevel(builder, level)


def MemberAddTid(builder, tid):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(tid), 0)


def AddTid(builder, tid):
    MemberAddTid(builder, tid)


def MemberStartTidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTidVector(builder, numElems: int) -> int:
    return MemberStartTidVector(builder, numElems)


def MemberAddSignature(builder, signature):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(signature), 0)


def AddSignature(builder, signature):
    MemberAddSignature(builder, signature)


def MemberStartSignatureVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartSignatureVector(builder, numElems: int) -> int:
    return MemberStartSignatureVector(builder, numElems)


def MemberEnd(builder):
    return builder.EndObject()


def End(builder):
    return MemberEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid

import numpy as np
from zlmdb import MapBytes20Uuid, MapStringUuid, MapUuidFlatBuffers, flatbuffers, table

from cfxdb import pack_uint256, unpack_uint256
from cfxdb.gen.xbrnetwork import Account as AccountGen


class _AccountGen(AccountGen.Account):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsAccount(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _AccountGen()
        x.Init(buf, n + offset)
        return x

    def OidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def WalletAddressAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def RegisteredAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Account(object):
    """
    XBR Network members.
    """

    ACCOUNT_LEVEL_NONE = 0
    """
    Unset.
    """

    ACCOUNT_LEVEL_ACTIVE = 1
    """
    Account is active.
    """

    ACCOUNT_LEVEL_VERIFIED = 2
    """
    Account is active and verified.
    """

    ACCOUNT_LEVEL_RETIRED = 3
    """
    Account is retired.
    """

    ACCOUNT_LEVEL_PENALTY = 4
    """
    Account is subject to a temporary penalty.
    """

    ACCOUNT_LEVEL_BLOCKED = 5
    """
    Account is currently blocked and cannot current actively participate in the market.
    """

    ACCOUNT_LEVEL = list(range(6))
    """
    Valid account levels.
    """

    WALLET_TYPE_NONE = 0
    """
    Wallet type unset ("null").
    """

    WALLET_TYPE_IMPORTED = 1
    """
    Account (primary) wallet was imported (the user provided the wallet public address).
    """

    WALLET_TYPE_METAMASK = 2
    """
    Account (primary) wallet was imported (the user provided the wallet public address).
    """

    WALLET_TYPE_HOSTED = 3
    """
    Account wallet in hosted (in this database).
    """

    WALLET_TYPE = list(range(4))
    """
    Valid account wallet types.
    """

    WALLET_TYPE_TO_STRING = {
        0: "none",
        1: "imported",
        2: "metamask",
        3: "hosted",
    }
    """
    Map of ``wallet-type-code`` to ``wallet-type-name``.
    """

    WALLET_TYPE_FROM_STRING = {
        "none": 0,
        "imported": 1,
        "metamask": 2,
        "hosted": 3,
    }
    """
    Map of ``wallet-type-name`` to ``wallet-type-code``.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # Globally unique and static member ID.
        # [uint8] (uuid)
        self._oid = None

        # Timestamp (epoch time in ns) of initial creation of this record.
        # uint64 (timestamp)
        self._created = None

        # XBR Network username (must be globall unique on https://xbr.network)
        # string
        self._username = None

        # User (primary) email address.
        # string
        self._email = None

        # Timestamp (epoch time in ns) when the user email was (last) verified or 0 if unverified.
        # uint64 (timestamp)
        self._email_verified = None

        # Type of (primary) user crypto wallet in use.
        # WalletType
        self._wallet_type = None

        # Public address of user crypto wallet in use.
        # [uint8] (address)
        self._wallet_address = None

        # Block number (on the blockchain) when the member (originally) registered.
        # [uint8] (uint256)
        self._registered = None

        # EULA the member agreed to when joining the market (IPFS Multihash string).
        # string (multihash)
        self._eula = None

        # Optional member profile (IPFS Multihash string).
        # string (multihash)
        self._profile = None

        # Current member level.
        # AccountLevel
        self._level = None

        self._recovery_algo = None
        self._recovery_data = None
        self._recovery_salt = None

    def marshal(self):
        obj = {
            "oid": self.oid.bytes if self.oid else None,
            "created": int(self.created) if self.created else None,
            "username": self.username,
            "email": self.email,
            "email_verified": int(self.email_verified) if self.email_verified is not None else None,
            "wallet_type": self.wallet_type,
            "wallet_address": bytes(self.wallet_address) if self.wallet_address else None,
            "registered": int(self.registered) if self.registered else None,
            "eula": self.eula,
            "profile": self.profile,
            "level": self.level,
            "recovery_algo": self.recovery_algo,
            "recovery_data": self.recovery_data,
            "recovery_salt": self.recovery_salt,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def recovery_algo(self) -> np.datetime64:
        """
        Timestamp (epoch time in ns) of initial creation of this record.
        """
        if self._created is None and self._from_fbs:
            self._created = np.datetime64(self._from_fbs.Created(), "ns")
        return self._created

    @recovery_algo.setter
    def recovery_algo(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._created = value

    @property
    def recovery_data(self) -> bytes:
        """
        Recovery backup data (usually encrypted according to the account backup recovery algo in use).
        """
        if self._recovery_data is None and self._from_fbs:
            if self._from_fbs.RecoveryDataLength():
                self._recovery_data = self._from_fbs.RecoveryDataAsBytes()
        return self._recovery_data

    @recovery_data.setter
    def recovery_data(self, value: bytes):
        assert value is None or type(value) == bytes
        self._recovery_data = value

    @property
    def recovery_salt(self) -> bytes:
        """
        Salt used for encryption of recovery data.
        """
        if self._recovery_salt is None and self._from_fbs:
            if self._from_fbs.RecoverySaltLength():
                self._recovery_salt = self._from_fbs.RecoverySaltAsBytes()
        return self._recovery_salt

    @recovery_salt.setter
    def recovery_salt(self, value: bytes):
        assert value is None or type(value) == bytes
        self._recovery_salt = value

    @property
    def oid(self) -> uuid.UUID:
        """
        Globally unique and static member ID.
        """
        if self._oid is None and self._from_fbs:
            if self._from_fbs.OidLength():
                _oid = self._from_fbs.OidAsBytes()
                self._oid = uuid.UUID(bytes=bytes(_oid))
            else:
                self._oid = uuid.UUID(bytes=b"\x00" * 20)
        return self._oid

    @oid.setter
    def oid(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._oid = value

    @property
    def created(self) -> np.datetime64:
        """
        Timestamp (epoch time in ns) of initial creation of this record.
        """
        if self._created is None and self._from_fbs:
            self._created = np.datetime64(self._from_fbs.Created(), "ns")
        return self._created

    @created.setter
    def created(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._created = value

    @property
    def username(self) -> str:
        """
        XBR Network username (must be globally unique on https://xbr.network)
        """
        if self._username is None and self._from_fbs:
            username = self._from_fbs.Username()
            if username:
                self._username = username.decode("utf8")
        return self._username

    @username.setter
    def username(self, value: str):
        assert value is None or type(value) == str
        self._username = value

    @property
    def email(self) -> str:
        """
        User (primary) email address.
        """
        if self._email is None and self._from_fbs:
            email = self._from_fbs.Email()
            if email:
                self._email = email.decode("utf8")
        return self._email

    @email.setter
    def email(self, value: str):
        assert value is None or type(value) == str
        self._email = value

    @property
    def email_verified(self) -> np.datetime64:
        """
        Timestamp (epoch time in ns) when the user email was (last) verified or 0 if unverified.
        """
        if self._email_verified is None and self._from_fbs:
            self._email_verified = np.datetime64(self._from_fbs.EmailVerified(), "ns")
        return self._email_verified

    @email_verified.setter
    def email_verified(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._email_verified = value

    @property
    def wallet_type(self) -> int:
        """
        Type of (primary) user crypto wallet in use.
        """
        if self._wallet_type is None and self._from_fbs:
            self._wallet_type = self._from_fbs.WalletType()
        return self._wallet_type or 0

    @wallet_type.setter
    def wallet_type(self, value: int):
        assert value is None or type(value) == int
        self._wallet_type = value

    @property
    def wallet_address(self) -> bytes:
        """
        Public address of user crypto wallet in use.
        """
        if self._wallet_address is None and self._from_fbs:
            if self._from_fbs.WalletAddressLength():
                self._wallet_address = self._from_fbs.WalletAddressAsBytes()
        return self._wallet_address

    @wallet_address.setter
    def wallet_address(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 20)
        self._wallet_address = value

    @property
    def registered(self) -> int:
        """
        Block number (on the blockchain) when the member (originally) registered.
        """
        if self._registered is None and self._from_fbs:
            if self._from_fbs.RegisteredLength():
                _registered = self._from_fbs.RegisteredAsBytes()
                self._registered = unpack_uint256(bytes(_registered))
            else:
                self._registered = 0
        return self._registered

    @registered.setter
    def registered(self, value: int):
        assert value is None or type(value) == int
        self._registered = value

    @property
    def eula(self) -> str:
        """
        EULA the member agreed to when joining the market (IPFS Multihash string).
        """
        if self._eula is None and self._from_fbs:
            eula = self._from_fbs.Eula()
            if eula:
                self._eula = eula.decode("utf8")
        return self._eula

    @eula.setter
    def eula(self, value: str):
        assert value is None or type(value) == str
        self._eula = value

    @property
    def profile(self) -> str:
        """
        Optional member profile (IPFS Multihash string).
        """
        if self._profile is None and self._from_fbs:
            profile = self._from_fbs.Profile()
            if profile:
                self._profile = profile.decode("utf8")
        return self._profile

    @profile.setter
    def profile(self, value: str):
        assert value is None or type(value) == str
        self._profile = value

    @property
    def level(self) -> int:
        """
        Current member level.
        """
        if self._level is None and self._from_fbs:
            self._level = self._from_fbs.Level()
        return self._level

    @level.setter
    def level(self, value: int):
        assert value is None or type(value) == int
        self._level = value

    @staticmethod
    def cast(buf):
        return Account(_AccountGen.GetRootAsAccount(buf, 0))

    def build(self, builder):
        oid = self.oid.bytes if self.oid else None
        if oid:
            oid = builder.CreateString(oid)

        username = self.username
        if username:
            username = builder.CreateString(username)

        email = self.email
        if email:
            email = builder.CreateString(email)

        wallet_address = self.wallet_address
        if wallet_address:
            wallet_address = builder.CreateString(wallet_address)

        registered = self.registered
        if registered:
            registered = builder.CreateString(pack_uint256(registered))

        eula = self.eula
        if eula:
            eula = builder.CreateString(eula)

        profile = self.profile
        if profile:
            profile = builder.CreateString(profile)

        AccountGen.AccountStart(builder)

        if oid:
            AccountGen.AccountAddOid(builder, oid)

        if self.created:
            AccountGen.AccountAddCreated(builder, int(self.created))

        if username:
            AccountGen.AccountAddUsername(builder, username)

        if email:
            AccountGen.AccountAddEmail(builder, email)

        if self.email_verified:
            AccountGen.AccountAddEmailVerified(builder, int(self.email_verified))

        if self.wallet_type:
            AccountGen.AccountAddWalletType(builder, self.wallet_type)

        if wallet_address:
            AccountGen.AccountAddWalletAddress(builder, wallet_address)

        if registered:
            AccountGen.AccountAddRegistered(builder, registered)

        if eula:
            AccountGen.AccountAddEula(builder, eula)

        if profile:
            AccountGen.AccountAddProfile(builder, profile)

        if self.level:
            AccountGen.AccountAddLevel(builder, self.level)

        final = AccountGen.AccountEnd(builder)

        return final


@table("d155dff2-ac36-4c69-b1b0-254ce2eb237d", build=Account.build, cast=Account.cast)
class Accounts(MapUuidFlatBuffers):
    """
    Database table for XBR member accounts.
    """

    @staticmethod
    def parse(data):
        """

        :param data:
        :return:
        """
        oid = None
        if "oid" in data:
            assert type(data["oid"] == bytes and len(data["oid"]) == 16)
            oid = uuid.UUID(bytes=data["oid"])

        created = None
        if "created" in data:
            assert type(data["created"] == int)
            created = np.datetime64(data["created"], "ns")

        username = None
        if "username" in data:
            assert type(data["username"] == str)
            username = data["username"]

        email = None
        if "email" in data:
            assert type(data["email"] == str)
            email = data["email"]

        email_verified = None
        if "email_verified" in data:
            assert type(data["email_verified"] == int)
            email_verified = np.datetime64(data["email_verified"], "ns")

        wallet_type = None
        if "wallet_type" in data:
            assert type(data["wallet_type"] == int)
            wallet_type = data["wallet_type"]

        wallet_address = None
        if "wallet_address" in data:
            assert type(data["wallet_address"] == bytes and len(data["wallet_address"]) == 20)
            wallet_address = data["wallet_address"]

        registered = None
        if "registered" in data:
            assert type(data["registered"] == int)
            registered = data["registered"]

        eula = None
        if "eula" in data:
            assert type(data["eula"] == str)
            eula = data["eula"]

        profile = None
        if "profile" in data:
            assert type(data["profile"] == str)
            profile = data["profile"]

        level = None
        if "level" in data:
            assert type(data["level"] == int)
            level = data["level"]

        obj = Account()
        obj.oid = oid
        obj.created = created
        obj.username = username
        obj.email = email
        obj.email_verified = email_verified
        obj.wallet_type = wallet_type
        obj.wallet_address = wallet_address
        obj.registered = registered
        obj.eula = eula
        obj.profile = profile
        obj.level = level

        return obj


@table("760d42a0-110e-474b-ab85-6e2396af035d")
class IndexAccountsByUsername(MapStringUuid):
    """
    Database (index) table for username-to-account mapping.
    """


@table("76706b89-8639-491e-8989-afc5a7c8a5b4")
class IndexAccountsByEmail(MapStringUuid):
    """
    Database (index) table for (user-)email-to-account mapping.
    """


@table("432ae4fa-a23e-45d7-b2a4-c9ae868df2b3")
class IndexAccountsByWallet(MapBytes20Uuid):
    """
    Database (index) table for (user-)wallet-to-account mapping.
    """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: realmstore

# The serializer the app payload was serialized with before encryption.
class EncSerializer(object):
    # Unset
    NONE = 0
    # Payload was serialized in JSON
    JSON = 1
    # Payload was serialized in MsgPack
    MSGPACK = 2
    # Payload was serialized in CBOR
    CBOR = 3
    # Payload was serialized in UBJSON
    UBJSON = 4
    OPAQUE = 5
    FLATBUFFERS = 6



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.log.mnode_logs import MNodeLogs
from cfxdb.log.mworker_logs import MWorkerLogs


class Schema(object):
    def __init__(self, db):
        self.db = db

    # mnode_logs: MNodeLogs
    mnode_logs = None
    """
    Managed node log records.
    """

    # mworker_logs: MWorkerLogs
    mworker_logs = None
    """
    Managed node worker log records.
    """

    @staticmethod
    def attach(db):
        """
        Factory to create a schema from attaching to a database. The schema tables
        will be automatically mapped as persistant maps and attached to the
        database slots.

        :param db: zlmdb.Database
        :return: object of Schema
        """
        schema = Schema(db)

        schema.mnode_logs = db.attach_table(MNodeLogs)
        schema.mworker_logs = db.attach_table(MWorkerLogs)

        return schema



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: arealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# WAMP authentication roles defined, including permissions.
class Role(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Role()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsRole(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Role
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ID of this object.
    # Role
    def Oid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Time when the object was created.
    # Role
    def Created(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Owner organization of this object.
    # Role
    def Owner(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Name of this role, must be unique within the management realm at any given point in time.
    # Role
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Set of static permissions on the role.
    # Role
    def Permissions(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 9
            from arealm.Permission import Permission

            obj = Permission()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Role
    def PermissionsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Role
    def PermissionsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # URIs or URI patterns the permissions must match.
    # Role
    def PermissionUris(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Role
    def PermissionUrisLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Role
    def PermissionUrisIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Alternatively to defining a static set of permissions, the WAMP procedure URI of a dynamic authorizer can be specified.
    # Role
    def Authorizer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def RoleStart(builder):
    builder.StartObject(7)


def Start(builder):
    RoleStart(builder)


def RoleAddOid(builder, oid):
    builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    RoleAddOid(builder, oid)


def RoleAddCreated(builder, created):
    builder.PrependUint64Slot(1, created, 0)


def AddCreated(builder, created):
    RoleAddCreated(builder, created)


def RoleAddOwner(builder, owner):
    builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(owner), 0)


def AddOwner(builder, owner):
    RoleAddOwner(builder, owner)


def RoleAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)


def AddName(builder, name):
    RoleAddName(builder, name)


def RoleAddPermissions(builder, permissions):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(permissions), 0)


def AddPermissions(builder, permissions):
    RoleAddPermissions(builder, permissions)


def RoleStartPermissionsVector(builder, numElems):
    return builder.StartVector(9, numElems, 1)


def StartPermissionsVector(builder, numElems: int) -> int:
    return RoleStartPermissionsVector(builder, numElems)


def RoleAddPermissionUris(builder, permissionUris):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(permissionUris), 0)


def AddPermissionUris(builder, permissionUris):
    RoleAddPermissionUris(builder, permissionUris)


def RoleStartPermissionUrisVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartPermissionUrisVector(builder, numElems: int) -> int:
    return RoleStartPermissionUrisVector(builder, numElems)


def RoleAddAuthorizer(builder, authorizer):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(authorizer), 0)


def AddAuthorizer(builder, authorizer):
    RoleAddAuthorizer(builder, authorizer)


def RoleEnd(builder):
    return builder.EndObject()


def End(builder):
    return RoleEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit
import uuid

import txaio

txaio.use_twisted()  # noqa

import numpy as np
import pytest
from txaio import time_ns
from zlmdb import flatbuffers

from cfxdb.tests._util import _gen_ipfs_hash
from cfxdb.xbr import Market


def fill_market(market):
    market.market = uuid.uuid4()
    market.timestamp = np.datetime64(time_ns(), "ns")
    market.seq = random.randint(1, 2**32 - 1)
    market.owner = os.urandom(20)
    market.coin = os.urandom(20)
    market.terms = _gen_ipfs_hash()
    market.meta = _gen_ipfs_hash()
    market.maker = os.urandom(20)
    market.provider_security = random.randint(0, 2**256 - 1)
    market.consumer_security = random.randint(0, 2**256 - 1)
    market.market_fee = random.randint(0, 2**256 - 1)
    market.tid = os.urandom(32)
    market.signature = os.urandom(65)


@pytest.fixture(scope="function")
def market():
    _market = Market()
    fill_market(_market)
    return _market


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def test_market_roundtrip(market, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = market.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 544

    # create python object from bytes (flatbuffes)
    _market = Market.cast(data)

    assert _market.market == market.market
    assert _market.timestamp == market.timestamp
    assert _market.seq == market.seq
    assert _market.owner == market.owner
    assert _market.coin == market.coin
    assert _market.terms == market.terms
    assert _market.meta == market.meta
    assert _market.maker == market.maker
    assert _market.provider_security == market.provider_security
    assert _market.consumer_security == market.consumer_security
    assert _market.market_fee == market.market_fee
    assert _market.tid == market.tid
    assert _market.signature == market.signature


def test_market_roundtrip_perf(market, builder):
    obj = market.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _market = Market.cast(data)
        if True:
            assert _market.market == market.market
            assert _market.timestamp == market.timestamp
            assert _market.seq == market.seq
            assert _market.owner == market.owner
            assert _market.coin == market.coin
            assert _market.terms == market.terms
            assert _market.meta == market.meta
            assert _market.maker == market.maker
            assert _market.provider_security == market.provider_security
            assert _market.consumer_security == market.consumer_security
            assert _market.market_fee == market.market_fee
            assert _market.tid == market.tid
            assert _market.signature == market.signature

            scratch["value"] += _market.seq

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################


class UserMrealmRoleFbs(object):
    """
    Database class for CFC user roles on a management realm using Flatbuffers.
    """

    def __init__(self):
        raise NotImplementedError("not yet implemented")



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import Any, Dict, List, Optional
from uuid import UUID

import numpy as np

from cfxdb.common import ConfigurationElement
from cfxdb.gen.arealm.ApplicationRealmStatus import ApplicationRealmStatus


class ApplicationRealm(ConfigurationElement):
    """
    Application realm database configuration object.
    """

    STATUS_BY_CODE = {
        ApplicationRealmStatus.NONE: "NONE",
        ApplicationRealmStatus.STOPPED: "STOPPED",
        ApplicationRealmStatus.STARTING: "STARTING",
        ApplicationRealmStatus.RUNNING: "RUNNING",
        ApplicationRealmStatus.PAUSED: "PAUSED",
        ApplicationRealmStatus.STOPPING: "STOPPING",
        ApplicationRealmStatus.ERROR: "ERROR",
        ApplicationRealmStatus.DEGRADED: "DEGRADED",
    }

    STATUS_BY_NAME = {
        "NONE": ApplicationRealmStatus.NONE,
        "STOPPED": ApplicationRealmStatus.STOPPED,
        "STARTING": ApplicationRealmStatus.STARTING,
        "RUNNING": ApplicationRealmStatus.RUNNING,
        "PAUSED": ApplicationRealmStatus.PAUSED,
        "STOPPING": ApplicationRealmStatus.STOPPING,
        "ERROR": ApplicationRealmStatus.ERROR,
        "DEGRADED": ApplicationRealmStatus.DEGRADED,
    }

    STATUS_STOPPED = ApplicationRealmStatus.STOPPED
    STATUS_STARTING = ApplicationRealmStatus.STARTING
    STATUS_RUNNING = ApplicationRealmStatus.RUNNING
    STATUS_PAUSED = ApplicationRealmStatus.PAUSED
    STATUS_STOPPING = ApplicationRealmStatus.STOPPING
    STATUS_ERROR = ApplicationRealmStatus.ERROR
    STATUS_DEGRADED = ApplicationRealmStatus.DEGRADED

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        name: Optional[str] = None,
        status: Optional[int] = None,
        workergroup_oid: Optional[UUID] = None,
        webcluster_oid: Optional[UUID] = None,
        datamarket_oid: Optional[UUID] = None,
        changed: Optional[np.datetime64] = None,
        owner_oid: Optional[UUID] = None,
        _unknown: Optional[Any] = None,
    ):
        """

        :param oid: Object ID of application realm

        :param label: Optional user label of application realm

        :param description: Optional user description of application realm

        :param tags: Optional list of user tags on application realm

        :param name: Name of application realm

        :param status: Status of application realm.

        :param workergroup_oid: When running, router cluster worker group this application
            realm is running on.

        :param webcluster_oid: When running, the web cluster to serve as a frontend layer
            for the application realm.

        :param datamarket_oid: When this application realm is to be federated with nodes paired
            to a different management realm (master node) or run by different operators.

        :param changed: Timestamp when the application realm was last changed

        :param owner_oid: Owning user (object ID)
        """
        ConfigurationElement.__init__(self, oid=oid, label=label, description=description, tags=tags)
        self.name = name
        self.status = status
        self.workergroup_oid = workergroup_oid
        self.webcluster_oid = webcluster_oid
        self.datamarket_oid = datamarket_oid
        self.changed = changed
        self.owner_oid = owner_oid

        # private member with unknown/untouched data passing through
        self._unknown = _unknown

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        if not ConfigurationElement.__eq__(self, other):
            return False
        if other.name != self.name:
            return False
        if other.status != self.status:
            return False
        if other.workergroup_oid != self.workergroup_oid:
            return False
        if other.webcluster_oid != self.webcluster_oid:
            return False
        if other.datamarket_oid != self.datamarket_oid:
            return False
        if other.changed != self.changed:
            return False
        if other.owner_oid != self.owner_oid:
            return False
        return True

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return pprint.pformat(self.marshal())

    def copy(self, other: "ApplicationRealm", overwrite: bool = False):
        """
        Copy over other object.

        :param other: Other application realm to copy data from.
        :param overwrite: Overwrite members already set.
        """
        ConfigurationElement.copy(self, other, overwrite=overwrite)

        if (not self.name and other.name) or overwrite:
            self.name = other.name
        if (not self.status and other.status) or overwrite:
            self.status = other.status
        if (not self.workergroup_oid and other.workergroup_oid) or overwrite:
            self.workergroup_oid = other.workergroup_oid
        if (not self.webcluster_oid and other.webcluster_oid) or overwrite:
            self.webcluster_oid = other.webcluster_oid
        if (not self.datamarket_oid and other.datamarket_oid) or overwrite:
            self.datamarket_oid = other.datamarket_oid
        if (not self.changed and other.changed) or overwrite:
            self.changed = other.changed
        if (not self.owner_oid and other.owner_oid) or overwrite:
            self.owner_oid = other.owner_oid

        # _unknown is not copied!

    def marshal(self) -> Dict[str, Any]:
        """
        Marshal this object to a generic host language object.
        """
        obj = ConfigurationElement.marshal(self)

        obj.update(
            {
                "oid": str(self.oid) if self.oid else None,
                "name": self.name,
                "status": ApplicationRealm.STATUS_BY_CODE[self.status]
                if self.status in ApplicationRealm.STATUS_BY_CODE
                else None,
                "workergroup_oid": str(self.workergroup_oid) if self.workergroup_oid else None,
                "webcluster_oid": str(self.webcluster_oid) if self.webcluster_oid else None,
                "datamarket_oid": str(self.datamarket_oid) if self.datamarket_oid else None,
                "changed": int(self.changed) if self.changed else None,
                "owner_oid": str(self.owner_oid) if self.owner_oid else None,
            }
        )

        if self._unknown:
            # pass through all attributes unknown
            obj.update(self._unknown)

        return obj

    @staticmethod
    def parse(data: Dict[str, Any]) -> "ApplicationRealm":
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :returns: New instance of this class.
        """
        assert type(data) == dict

        obj = ConfigurationElement.parse(data)
        data = obj._unknown

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["oid", "name", "status", "workergroup_oid", "webcluster_oid", "owner_oid", "changed"]:
                _unknown[k] = data[k]

        name = data.get("name", "arealm-{}".format(str(obj.oid)[:8]))
        assert type(name) == str

        status = data.get("status", None)
        assert status is None or (type(status) == str)
        status = ApplicationRealm.STATUS_BY_NAME[status] if status in ApplicationRealm.STATUS_BY_NAME else None

        workergroup_oid = None
        if "workergroup_oid" in data and data["workergroup_oid"] is not None:
            assert type(data["workergroup_oid"]) == str, "workergroup_oid must be a string, but was {}".format(
                type(data["workergroup_oid"])
            )
            workergroup_oid = UUID(data["workergroup_oid"])

        webcluster_oid = None
        if "webcluster_oid" in data and data["webcluster_oid"] is not None:
            assert type(data["webcluster_oid"]) == str, "webcluster_oid must be a string, but was {}".format(
                type(data["webcluster_oid"])
            )
            webcluster_oid = UUID(data["webcluster_oid"])

        datamarket_oid = None
        if "datamarket_oid" in data and data["datamarket_oid"] is not None:
            assert type(data["datamarket_oid"]) == str, "datamarket_oid must be a string, but was {}".format(
                type(data["datamarket_oid"])
            )
            datamarket_oid = UUID(data["datamarket_oid"])

        owner_oid = None
        if "owner_oid" in data and data["owner_oid"] is not None:
            assert type(data["owner_oid"]) == str, "owner_oid must be a string, but was {}".format(
                type(data["owner_oid"])
            )
            owner_oid = UUID(data["owner_oid"])

        changed = data.get("changed", None)
        assert changed is None or type(changed) == int
        if changed:
            changed = np.datetime64(changed, "ns")

        obj = ApplicationRealm(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            name=name,
            workergroup_oid=workergroup_oid,
            webcluster_oid=webcluster_oid,
            datamarket_oid=datamarket_oid,
            status=status,
            owner_oid=owner_oid,
            changed=changed,
            _unknown=_unknown,
        )

        return obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: arealm

# WAMP authentication methods.
class AuthenticationMethod(object):
    # No authentiation method defined (undefined).
    NONE = 0
    # WAMP-anonymous authentication.
    ANONYMOUS = 1
    # WAMP-ticket authentication.
    TICKET = 2
    # WAMP-CRA authentication.
    WAMPCRA = 3
    # TLS client certificate authentication [TRANSPORT LEVEL].
    TLS = 4
    # HTTP cookie authentication [TRANSPORT LEVEL].
    COOKIE = 5
    # WAMP-cryptosign authentication.
    CRYPTOSIGN = 6
    # WAMP-SCRAM authentication.
    SCRAM = 7



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.gen.arealm.ApplicationRealmStatus import ApplicationRealmStatus
from cfxdb.gen.arealm.AuthenticationMethod import AuthenticationMethod
from cfxdb.gen.arealm.MatchType import MatchType
from cfxdb.gen.arealm.UriCheckLevel import UriCheckLevel
from cfxdb.gen.mrealm.ClusterStatus import ClusterStatus
from cfxdb.gen.mrealm.WorkerGroupStatus import WorkerGroupStatus
from cfxdb.mrealm.application_realm import ApplicationRealm
from cfxdb.mrealm.arealm_role_association import ApplicationRealmRoleAssociation
from cfxdb.mrealm.cluster import Cluster
from cfxdb.mrealm.cluster_node_membership import ClusterNodeMembership
from cfxdb.mrealm.credential import Credential
from cfxdb.mrealm.management_realm import ManagementRealm
from cfxdb.mrealm.node import Node
from cfxdb.mrealm.permission import Permission
from cfxdb.mrealm.principal import Principal
from cfxdb.mrealm.role import Role
from cfxdb.mrealm.router_cluster import RouterCluster
from cfxdb.mrealm.router_cluster_node_membership import RouterClusterNodeMembership
from cfxdb.mrealm.router_workergroup import RouterWorkerGroup
from cfxdb.mrealm.router_workergroup_cluster_placement import RouterWorkerGroupClusterPlacement
from cfxdb.mrealm.web_cluster import WebCluster
from cfxdb.mrealm.web_cluster_node_membership import WebClusterNodeMembership
from cfxdb.mrealm.web_service import WebService

__all__ = (
    "ManagementRealm",
    "Node",
    "Cluster",
    "RouterCluster",
    "WebCluster",
    "ClusterNodeMembership",
    "WebClusterNodeMembership",
    "RouterClusterNodeMembership",
    "WebService",
    "ClusterStatus",
    "WorkerGroupStatus",
    "RouterWorkerGroup",
    "RouterWorkerGroupClusterPlacement",
    "MatchType",
    "UriCheckLevel",
    "Role",
    "Permission",
    "ApplicationRealm",
    "ApplicationRealmRoleAssociation",
    "ApplicationRealmStatus",
    "AuthenticationMethod",
    "Credential",
    "Principal",
)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

# monkey patch eth_abi for master branch (which we need for python 3.11)
# https://github.com/ethereum/eth-abi/blob/master/docs/release_notes.rst#breaking-changes
# https://github.com/ethereum/eth-abi/pull/161
# ImportError: cannot import name 'encode_single' from 'eth_abi' (/home/oberstet/cpy311_2/lib/python3.11/site-packages/eth_abi/__init__.py)
import eth_abi

if not hasattr(eth_abi, "encode_abi") and hasattr(eth_abi, "encode"):
    eth_abi.encode_abi = eth_abi.encode
if not hasattr(eth_abi, "encode_single") and hasattr(eth_abi, "encode"):
    eth_abi.encode_single = eth_abi.encode

# monkey patch web3 for master branch / upcoming v6 (which we need for python 3.11)
# AttributeError: type object 'Web3' has no attribute 'toChecksumAddress'. Did you mean: 'to_checksum_address'?
import web3

if not hasattr(web3.Web3, "toChecksumAddress") and hasattr(web3.Web3, "to_checksum_address"):
    web3.Web3.toChecksumAddress = web3.Web3.to_checksum_address
if not hasattr(web3.Web3, "isConnected") and hasattr(web3.Web3, "is_connected"):
    web3.Web3.isConnected = web3.Web3.is_connected
if not hasattr(web3.Web3, "toBytes") and hasattr(web3.Web3, "to_bytes"):
    web3.Web3.toBytes = web3.Web3.to_bytes
if not hasattr(web3.Web3, "toInt") and hasattr(web3.Web3, "to_int"):
    web3.Web3.toInt = web3.Web3.to_int

import sys

import txaio

txaio.use_twisted()

# =============================================================================
# Monkey-patch zlmdb to expose vendored flatbuffers
# TEMPORARY: Until zlmdb 25.12.2 is released with native support for:
#   - `from zlmdb import flatbuffers`
#   - `zlmdb.setup_flatbuffers_import()`
# See: https://github.com/crossbario/zlmdb/issues/XX (TODO: add issue number)
# =============================================================================
import zlmdb

# Expose zlmdb's vendored flatbuffers as zlmdb.flatbuffers
if not hasattr(zlmdb, "flatbuffers"):
    zlmdb.flatbuffers = zlmdb._flatbuffers_vendor
    sys.modules.setdefault("zlmdb.flatbuffers", zlmdb._flatbuffers_vendor)


def _setup_flatbuffers_import():
    """
    Register zlmdb's vendored flatbuffers in sys.modules.

    This allows generated flatbuffers code (which does `import flatbuffers`)
    to resolve to zlmdb's vendored copy.
    """
    _vendor = zlmdb._flatbuffers_vendor
    sys.modules.setdefault("flatbuffers", _vendor)
    sys.modules.setdefault("flatbuffers.compat", _vendor.compat)
    sys.modules.setdefault("flatbuffers.builder", _vendor.builder)
    sys.modules.setdefault("flatbuffers.table", _vendor.table)
    sys.modules.setdefault("flatbuffers.util", _vendor.util)
    sys.modules.setdefault("flatbuffers.number_types", _vendor.number_types)
    sys.modules.setdefault("flatbuffers.packer", _vendor.packer)
    sys.modules.setdefault("flatbuffers.encode", _vendor.encode)


if not hasattr(zlmdb, "setup_flatbuffers_import"):
    zlmdb.setup_flatbuffers_import = _setup_flatbuffers_import

# =============================================================================

from ._version import __version__  # noqa
from ._exception import InvalidConfigException  # noqa
from .common import address  # noqa
from .common import uint256, unpack_uint256, pack_uint256  # noqa
from .common import uint128, unpack_uint128, pack_uint128  # noqa
from .common import uint64, unpack_uint64, pack_uint64  # noqa
from . import meta, mrealm, xbr, xbrmm, xbrnetwork  # noqa
from . import globalschema, mrealmschema, cookiestore, realmstore  # noqa

__all__ = (
    "__version__",
    "meta",
    "mrealm",
    "cookiestore",
    "realmstore",
    "xbr",
    "xbrmm",
    "xbrnetwork",
    "address",
    "uint256",
    "pack_uint256",
    "unpack_uint256",
    "uint128",
    "pack_uint128",
    "unpack_uint128",
    "uint64",
    "pack_uint64",
    "unpack_uint64",
    "globalschema",
    "mrealmschema",
    "InvalidConfigException",
)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: arealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Association (N:M) between roles and application realms, both defined independently at the management realm level.
class ApplicationRealmRole(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ApplicationRealmRole()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsApplicationRealmRole(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # ApplicationRealmRole
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ID of the role being associated with an application realm.
    # ApplicationRealmRole
    def RoleOid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ID of the application realm this role is associated with.
    # ApplicationRealmRole
    def ArealmOid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None


def ApplicationRealmRoleStart(builder):
    builder.StartObject(2)


def Start(builder):
    ApplicationRealmRoleStart(builder)


def ApplicationRealmRoleAddRoleOid(builder, roleOid):
    builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(roleOid), 0)


def AddRoleOid(builder, roleOid):
    ApplicationRealmRoleAddRoleOid(builder, roleOid)


def ApplicationRealmRoleAddArealmOid(builder, arealmOid):
    builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(arealmOid), 0)


def AddArealmOid(builder, arealmOid):
    ApplicationRealmRoleAddArealmOid(builder, arealmOid)


def ApplicationRealmRoleEnd(builder):
    return builder.EndObject()


def End(builder):
    return ApplicationRealmRoleEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import uuid
from datetime import datetime
from pprint import pformat

from cfxdb.gen.user.ActivationStatus import ActivationStatus
from cfxdb.gen.user.ActivationType import ActivationType


class ActivationToken(object):
    """
    CFC user activation token database class for CBOR.
    """

    def __init__(
        self,
        oid=None,
        atype=None,
        status=None,
        created=None,
        completed=None,
        code=None,
        email=None,
        pubkey=None,
        _unknown=None,
    ):
        self.oid = oid
        self.atype = atype
        self.status = status
        self.created = created
        self.completed = completed
        self.code = code
        self.email = email
        self.pubkey = pubkey

        self._unknown = _unknown

    def __str__(self):
        return pformat(self.marshal())

    def copy(self, other):
        self.oid = other.oid
        self.atype = other.atype
        self.status = other.status
        self.created = other.created
        self.completed = other.completed
        self.code = other.code
        self.email = other.email
        self.pubkey = other.pubkey

        # _unknown is not copied!

    def marshal(self):
        assert isinstance(self.oid, uuid.UUID)
        assert self.atype in [ActivationType.LOGIN, ActivationType.REGISTRATION]
        assert self.status in [ActivationStatus.EXPIRED, ActivationStatus.PENDING, ActivationStatus.ACTIVE]
        assert isinstance(self.created, datetime)
        assert self.completed is None or isinstance(self.completed, datetime)
        assert self.code is None or type(self.code) == str
        assert self.email is None or type(self.email) == str
        assert self.pubkey is None or (type(self.pubkey) == str and len(self.pubkey) == 64)
        assert self._unknown is None or type(self._unknown) == dict

        created = float(self.created.timestamp()) * 1000000.0
        completed = None
        if self.completed:
            completed = float(self.completed.timestamp()) * 1000000.0

        obj = {
            "oid": str(self.oid),
            "atype": self.atype,
            "status": self.status,
            "created": created,
            "completed": completed,
            "code": self.code,
            "email": self.email,
            "pubkey": self.pubkey,
        }

        if self._unknown:
            # pass through all attributes unknown
            obj.update(self._unknown)
        else:
            return obj

    @staticmethod
    def parse(data):
        assert type(data) == dict

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["oid", "atype", "status", "created", "completed", "code", "email", "pubkey"]:
                val = data.pop(k)
                _unknown[k] = val

        oid = data.get("oid", None)
        assert type(oid) == str
        oid = uuid.UUID(oid)

        atype = data.get("atype", None)
        assert type(atype) == int

        status = data.get("status", None)
        assert type(status) == int

        created = data.get("created", None)
        assert type(created) == float or type(created) == int

        created = datetime.fromtimestamp(float(created) / 1000000.0)
        # created = datetime.utcfromtimestamp(float(created) / 1000000.)

        completed = data.get("completed", None)
        assert completed is None or type(completed) == float or type(completed) == int
        if completed:
            # https://docs.python.org
            # /3/library/time.html#time.time_ns
            # https://docs.python.org/3/library/datetime.html#datetime.datetime.timestamp

            completed = datetime.fromtimestamp(float(completed) / 1000000.0)
            # completed = datetime.utcfromtimestamp(float(completed) / 1000000.)

        code = data.get("code", None)
        assert type(code) == str

        email = data.get("email", None)
        assert email is None or type(email) == str

        pubkey = data.get("pubkey", None)
        assert pubkey is None or type(pubkey) == str
        if pubkey:
            assert len(pubkey) == 64

        obj = ActivationToken(
            oid=oid,
            atype=atype,
            status=status,
            created=created,
            completed=completed,
            code=code,
            email=email,
            pubkey=pubkey,
            _unknown=_unknown,
        )
        return obj



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from zlmdb import MapTimestampUuidFlatBuffers, table

from cfxdb.log.mnode_log import MNodeLog


@table("256a071f-5aeb-47f3-8786-97cd8281bdb7", build=MNodeLog.build, cast=MNodeLog.cast)
class MNodeLogs(MapTimestampUuidFlatBuffers):
    """
    Managed node heartbeat log records (``(timestamp, node_id) -> node_log``).
    """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: cookiestore

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Persistent cookies, as used in WAMP-Cookie authentication by router and proxy workers.
class Cookie(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Cookie()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsCookie(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Cookie
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Database ID of this cookie record.
    # Cookie
    def Oid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Cookie
    def OidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Cookie
    def OidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Cookie
    def OidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Timestamp when the cookie was created. Epoch time in ns.
    # Cookie
    def Created(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Cookie maximum age (lifetime of the cookie in seconds, see http://tools.ietf.org/html/rfc6265#page-20), e.g. ``604800"``.
    # Cookie
    def MaxAge(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Cookie name, as set in HTTP header, e.g. ``"cbtid"``.
    # Cookie
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Cookie value, as set in HTTP header, e.g. ``"gn2ri8fuAYQse50/L6N7jnt2"``.
    # Cookie
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Timestamp when the cookie was authenticated (if any). Epoch time in ns.
    # Cookie
    def Authenticated(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # The Crossbar.io node (within the management domain) the cookie was authenticated on (if any).
    # Cookie
    def AuthenticatedOnNode(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Cookie
    def AuthenticatedOnNodeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Cookie
    def AuthenticatedOnNodeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Cookie
    def AuthenticatedOnNodeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # The Crossbar.io worker (within the node) the cookie was authenticated on (if any).
    # Cookie
    def AuthenticatedOnWorker(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The client transport information for the connection the cookie was authenticated in (if any).
    # Cookie
    def AuthenticatedTransportInfo(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Cookie
    def AuthenticatedTransportInfoAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Cookie
    def AuthenticatedTransportInfoLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Cookie
    def AuthenticatedTransportInfoIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        return o == 0

    # The WAMP session ID of the original authenticating session.
    # Cookie
    def AuthenticatedSession(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Timestamp when the original authenticating session was welcome by the router. Epoch time in ns.
    # Cookie
    def AuthenticatedJoinedAt(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # The (original) WAMP authentication method, after which the client was authenticated before setting this cookie.
    # Cookie
    def AuthenticatedAuthmethod(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP authid a cookie-authenticating session is to be assigned.
    # Cookie
    def Authid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP authrole a cookie-authenticating session is to join under.
    # Cookie
    def Authrole(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP authmethod a cookie-authenticating session is to join under.
    # Cookie
    def Authmethod(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP realm a cookie-authenticating session is to join.
    # Cookie
    def Authrealm(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # The WAMP authentication extra data to be returned to the client performing cookie-based authentication.
    # Cookie
    def Authextra(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Cookie
    def AuthextraAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Cookie
    def AuthextraLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Cookie
    def AuthextraIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        return o == 0


def CookieStart(builder):
    builder.StartObject(17)


def Start(builder):
    CookieStart(builder)


def CookieAddOid(builder, oid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    CookieAddOid(builder, oid)


def CookieStartOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOidVector(builder, numElems: int) -> int:
    return CookieStartOidVector(builder, numElems)


def CookieAddCreated(builder, created):
    builder.PrependUint64Slot(1, created, 0)


def AddCreated(builder, created):
    CookieAddCreated(builder, created)


def CookieAddMaxAge(builder, maxAge):
    builder.PrependUint64Slot(2, maxAge, 0)


def AddMaxAge(builder, maxAge):
    CookieAddMaxAge(builder, maxAge)


def CookieAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)


def AddName(builder, name):
    CookieAddName(builder, name)


def CookieAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0)


def AddValue(builder, value):
    CookieAddValue(builder, value)


def CookieAddAuthenticated(builder, authenticated):
    builder.PrependUint64Slot(5, authenticated, 0)


def AddAuthenticated(builder, authenticated):
    CookieAddAuthenticated(builder, authenticated)


def CookieAddAuthenticatedOnNode(builder, authenticatedOnNode):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(authenticatedOnNode), 0)


def AddAuthenticatedOnNode(builder, authenticatedOnNode):
    CookieAddAuthenticatedOnNode(builder, authenticatedOnNode)


def CookieStartAuthenticatedOnNodeVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAuthenticatedOnNodeVector(builder, numElems: int) -> int:
    return CookieStartAuthenticatedOnNodeVector(builder, numElems)


def CookieAddAuthenticatedOnWorker(builder, authenticatedOnWorker):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(authenticatedOnWorker), 0)


def AddAuthenticatedOnWorker(builder, authenticatedOnWorker):
    CookieAddAuthenticatedOnWorker(builder, authenticatedOnWorker)


def CookieAddAuthenticatedTransportInfo(builder, authenticatedTransportInfo):
    builder.PrependUOffsetTRelativeSlot(
        8, flatbuffers.number_types.UOffsetTFlags.py_type(authenticatedTransportInfo), 0
    )


def AddAuthenticatedTransportInfo(builder, authenticatedTransportInfo):
    CookieAddAuthenticatedTransportInfo(builder, authenticatedTransportInfo)


def CookieStartAuthenticatedTransportInfoVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAuthenticatedTransportInfoVector(builder, numElems: int) -> int:
    return CookieStartAuthenticatedTransportInfoVector(builder, numElems)


def CookieAddAuthenticatedSession(builder, authenticatedSession):
    builder.PrependUint64Slot(9, authenticatedSession, 0)


def AddAuthenticatedSession(builder, authenticatedSession):
    CookieAddAuthenticatedSession(builder, authenticatedSession)


def CookieAddAuthenticatedJoinedAt(builder, authenticatedJoinedAt):
    builder.PrependUint64Slot(10, authenticatedJoinedAt, 0)


def AddAuthenticatedJoinedAt(builder, authenticatedJoinedAt):
    CookieAddAuthenticatedJoinedAt(builder, authenticatedJoinedAt)


def CookieAddAuthenticatedAuthmethod(builder, authenticatedAuthmethod):
    builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(authenticatedAuthmethod), 0)


def AddAuthenticatedAuthmethod(builder, authenticatedAuthmethod):
    CookieAddAuthenticatedAuthmethod(builder, authenticatedAuthmethod)


def CookieAddAuthid(builder, authid):
    builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(authid), 0)


def AddAuthid(builder, authid):
    CookieAddAuthid(builder, authid)


def CookieAddAuthrole(builder, authrole):
    builder.PrependUOffsetTRelativeSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(authrole), 0)


def AddAuthrole(builder, authrole):
    CookieAddAuthrole(builder, authrole)


def CookieAddAuthmethod(builder, authmethod):
    builder.PrependUOffsetTRelativeSlot(14, flatbuffers.number_types.UOffsetTFlags.py_type(authmethod), 0)


def AddAuthmethod(builder, authmethod):
    CookieAddAuthmethod(builder, authmethod)


def CookieAddAuthrealm(builder, authrealm):
    builder.PrependUOffsetTRelativeSlot(15, flatbuffers.number_types.UOffsetTFlags.py_type(authrealm), 0)


def AddAuthrealm(builder, authrealm):
    CookieAddAuthrealm(builder, authrealm)


def CookieAddAuthextra(builder, authextra):
    builder.PrependUOffsetTRelativeSlot(16, flatbuffers.number_types.UOffsetTFlags.py_type(authextra), 0)


def AddAuthextra(builder, authextra):
    CookieAddAuthextra(builder, authextra)


def CookieStartAuthextraVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAuthextraVector(builder, numElems: int) -> int:
    return CookieStartAuthextraVector(builder, numElems)


def CookieEnd(builder):
    return builder.EndObject()


def End(builder):
    return CookieEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: log

# Managed node worker types.
class MWorkerType(object):
    # Unset worker type.
    NONE = 0
    # Router worker.
    ROUTER = 1
    # Router worker.
    CONTAINER = 2
    # Router worker.
    GUEST = 3
    # Router worker.
    TESTEE = 4
    # Router worker.
    PROXY = 5
    # Router worker.
    XBRMM = 6
    # Router worker.
    HOSTMONITOR = 7
    # Router worker.
    CONTROLLER = 8



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrmm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# XBR channel (from XBR consumer to XBR market maker) and XBR paying channels (from XBR market maker to XBR provider).
class Channel(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Channel()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsChannel(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Channel
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # OID of the market this payment channel is associated with.
    # Channel
    def MarketOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def MarketOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def MarketOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def MarketOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # OID of the buyer or seller actor in the market (a XBR network member) this channel is associated with.
    # Channel
    def MemberOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def MemberOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def MemberOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def MemberOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # OID of the channel.
    # Channel
    def ChannelOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def ChannelOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def ChannelOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def ChannelOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Database transaction time (epoch time in ns).
    # Channel
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Block number (on the blockchain) when the channel was opened.
    # Channel
    def OpenAt(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def OpenAtAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def OpenAtLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def OpenAtIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Channel sequence number.
    # Channel
    def Seq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Channel type: payment channel (from XBR consumer to XBR market maker) or paying channel (from XBR market maker to XBR provider).
    # Channel
    def ChannelType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Ethereum address of the sender delegate (either XBR Consumer delegate or XBR Market delegate == market maker)
    # Channel
    def Marketmaker(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def MarketmakerAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def MarketmakerLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def MarketmakerIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # Ethereum address of the sender (either XBR Consumer or XBR Market).
    # Channel
    def Actor(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def ActorAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def ActorLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def ActorIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        return o == 0

    # Ethereum address of the sender delegate (either XBR Consumer delegate or XBR Market delegate == market maker)
    # Channel
    def Delegate(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def DelegateAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def DelegateLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def DelegateIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        return o == 0

    # Ethereum address of the recipient (either XBR Market or XBR Provider)
    # Channel
    def Recipient(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def RecipientAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def RecipientLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def RecipientIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        return o == 0

    # Amount of XBR tokens initially deposited into the channel.
    # Channel
    def Amount(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def AmountAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def AmountLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def AmountIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        return o == 0

    # Channel (non-cooperative) close timeout in blocks (committed on the blockchain).
    # Channel
    def Timeout(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Current state of channel.
    # Channel
    def State(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Block number (on the blockchain) when the channel will close (at the latest).
    # Channel
    def ClosingAt(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def ClosingAtAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def ClosingAtLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def ClosingAtIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        return o == 0

    # Block number (on the blockchain) when the channel was finally closed on-chain.
    # Channel
    def ClosedAt(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def ClosedAtAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def ClosedAtLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def ClosedAtIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        return o == 0

    # Closing signature by market maker.
    # Channel
    def CloseMmSig(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def CloseMmSigAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def CloseMmSigLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def CloseMmSigIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        return o == 0

    # Closing signature by (seller or buyer) delegate.
    # Channel
    def CloseDelSig(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def CloseDelSigAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def CloseDelSigLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def CloseDelSigIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        return o == 0

    # Last off-chain, closing transaction: channel transaction sequence number.
    # Channel
    def CloseChannelSeq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Remaining (closing) channel balance (XBR).
    # Channel
    def CloseBalance(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def CloseBalanceAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def CloseBalanceLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def CloseBalanceIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        return o == 0

    # Flag indication if close is final (happens immediately without a channel timeout).
    # Channel
    def CloseIsFinal(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # When channel was finally closed on-chain, the Ethereum transaction ID.
    # Channel
    def ClosedTx(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Channel
    def ClosedTxAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Channel
    def ClosedTxLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Channel
    def ClosedTxIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        return o == 0


def ChannelStart(builder):
    builder.StartObject(22)


def Start(builder):
    ChannelStart(builder)


def ChannelAddMarketOid(builder, marketOid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(marketOid), 0)


def AddMarketOid(builder, marketOid):
    ChannelAddMarketOid(builder, marketOid)


def ChannelStartMarketOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMarketOidVector(builder, numElems: int) -> int:
    return ChannelStartMarketOidVector(builder, numElems)


def ChannelAddMemberOid(builder, memberOid):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(memberOid), 0)


def AddMemberOid(builder, memberOid):
    ChannelAddMemberOid(builder, memberOid)


def ChannelStartMemberOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMemberOidVector(builder, numElems: int) -> int:
    return ChannelStartMemberOidVector(builder, numElems)


def ChannelAddChannelOid(builder, channelOid):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(channelOid), 0)


def AddChannelOid(builder, channelOid):
    ChannelAddChannelOid(builder, channelOid)


def ChannelStartChannelOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartChannelOidVector(builder, numElems: int) -> int:
    return ChannelStartChannelOidVector(builder, numElems)


def ChannelAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(3, timestamp, 0)


def AddTimestamp(builder, timestamp):
    ChannelAddTimestamp(builder, timestamp)


def ChannelAddOpenAt(builder, openAt):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(openAt), 0)


def AddOpenAt(builder, openAt):
    ChannelAddOpenAt(builder, openAt)


def ChannelStartOpenAtVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOpenAtVector(builder, numElems: int) -> int:
    return ChannelStartOpenAtVector(builder, numElems)


def ChannelAddSeq(builder, seq):
    builder.PrependUint32Slot(5, seq, 0)


def AddSeq(builder, seq):
    ChannelAddSeq(builder, seq)


def ChannelAddChannelType(builder, channelType):
    builder.PrependUint8Slot(6, channelType, 0)


def AddChannelType(builder, channelType):
    ChannelAddChannelType(builder, channelType)


def ChannelAddMarketmaker(builder, marketmaker):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(marketmaker), 0)


def AddMarketmaker(builder, marketmaker):
    ChannelAddMarketmaker(builder, marketmaker)


def ChannelStartMarketmakerVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartMarketmakerVector(builder, numElems: int) -> int:
    return ChannelStartMarketmakerVector(builder, numElems)


def ChannelAddActor(builder, actor):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(actor), 0)


def AddActor(builder, actor):
    ChannelAddActor(builder, actor)


def ChannelStartActorVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartActorVector(builder, numElems: int) -> int:
    return ChannelStartActorVector(builder, numElems)


def ChannelAddDelegate(builder, delegate):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(delegate), 0)


def AddDelegate(builder, delegate):
    ChannelAddDelegate(builder, delegate)


def ChannelStartDelegateVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartDelegateVector(builder, numElems: int) -> int:
    return ChannelStartDelegateVector(builder, numElems)


def ChannelAddRecipient(builder, recipient):
    builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(recipient), 0)


def AddRecipient(builder, recipient):
    ChannelAddRecipient(builder, recipient)


def ChannelStartRecipientVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartRecipientVector(builder, numElems: int) -> int:
    return ChannelStartRecipientVector(builder, numElems)


def ChannelAddAmount(builder, amount):
    builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(amount), 0)


def AddAmount(builder, amount):
    ChannelAddAmount(builder, amount)


def ChannelStartAmountVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartAmountVector(builder, numElems: int) -> int:
    return ChannelStartAmountVector(builder, numElems)


def ChannelAddTimeout(builder, timeout):
    builder.PrependUint32Slot(12, timeout, 0)


def AddTimeout(builder, timeout):
    ChannelAddTimeout(builder, timeout)


def ChannelAddState(builder, state):
    builder.PrependUint8Slot(13, state, 0)


def AddState(builder, state):
    ChannelAddState(builder, state)


def ChannelAddClosingAt(builder, closingAt):
    builder.PrependUOffsetTRelativeSlot(14, flatbuffers.number_types.UOffsetTFlags.py_type(closingAt), 0)


def AddClosingAt(builder, closingAt):
    ChannelAddClosingAt(builder, closingAt)


def ChannelStartClosingAtVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartClosingAtVector(builder, numElems: int) -> int:
    return ChannelStartClosingAtVector(builder, numElems)


def ChannelAddClosedAt(builder, closedAt):
    builder.PrependUOffsetTRelativeSlot(15, flatbuffers.number_types.UOffsetTFlags.py_type(closedAt), 0)


def AddClosedAt(builder, closedAt):
    ChannelAddClosedAt(builder, closedAt)


def ChannelStartClosedAtVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartClosedAtVector(builder, numElems: int) -> int:
    return ChannelStartClosedAtVector(builder, numElems)


def ChannelAddCloseMmSig(builder, closeMmSig):
    builder.PrependUOffsetTRelativeSlot(16, flatbuffers.number_types.UOffsetTFlags.py_type(closeMmSig), 0)


def AddCloseMmSig(builder, closeMmSig):
    ChannelAddCloseMmSig(builder, closeMmSig)


def ChannelStartCloseMmSigVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartCloseMmSigVector(builder, numElems: int) -> int:
    return ChannelStartCloseMmSigVector(builder, numElems)


def ChannelAddCloseDelSig(builder, closeDelSig):
    builder.PrependUOffsetTRelativeSlot(17, flatbuffers.number_types.UOffsetTFlags.py_type(closeDelSig), 0)


def AddCloseDelSig(builder, closeDelSig):
    ChannelAddCloseDelSig(builder, closeDelSig)


def ChannelStartCloseDelSigVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartCloseDelSigVector(builder, numElems: int) -> int:
    return ChannelStartCloseDelSigVector(builder, numElems)


def ChannelAddCloseChannelSeq(builder, closeChannelSeq):
    builder.PrependUint32Slot(18, closeChannelSeq, 0)


def AddCloseChannelSeq(builder, closeChannelSeq):
    ChannelAddCloseChannelSeq(builder, closeChannelSeq)


def ChannelAddCloseBalance(builder, closeBalance):
    builder.PrependUOffsetTRelativeSlot(19, flatbuffers.number_types.UOffsetTFlags.py_type(closeBalance), 0)


def AddCloseBalance(builder, closeBalance):
    ChannelAddCloseBalance(builder, closeBalance)


def ChannelStartCloseBalanceVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartCloseBalanceVector(builder, numElems: int) -> int:
    return ChannelStartCloseBalanceVector(builder, numElems)


def ChannelAddCloseIsFinal(builder, closeIsFinal):
    builder.PrependBoolSlot(20, closeIsFinal, 0)


def AddCloseIsFinal(builder, closeIsFinal):
    ChannelAddCloseIsFinal(builder, closeIsFinal)


def ChannelAddClosedTx(builder, closedTx):
    builder.PrependUOffsetTRelativeSlot(21, flatbuffers.number_types.UOffsetTFlags.py_type(closedTx), 0)


def AddClosedTx(builder, closedTx):
    ChannelAddClosedTx(builder, closedTx)


def ChannelStartClosedTxVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartClosedTxVector(builder, numElems: int) -> int:
    return ChannelStartClosedTxVector(builder, numElems)


def ChannelEnd(builder):
    return builder.EndObject()


def End(builder):
    return ChannelEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import platform
import random
import timeit
import uuid

import pytest
import zlmdb
from autobahn import util
from txaio import (
    time_ns,
    with_twisted,  # noqa
)
from zlmdb import flatbuffers

from cfxdb.realmstore import Publication

zlmdb.TABLES_BY_UUID = {}


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def fill_publication(publication):
    publication.timestamp = time_ns()
    publication.publication = util.id()
    publication.publisher = util.id()
    publication.topic = "com.example.foobar.{}.doit".format(uuid.uuid4())

    publication.args = [23, "hello", {"foo": 0.5}]
    publication.kwargs = {"bar": 23, "baz": [1, 2, 3]}
    publication.payload = os.urandom(32)

    publication.acknowledge = random.choice([True, False])
    publication.retain = random.choice([True, False])
    publication.exclude_me = random.choice([True, False])

    i0 = util.id()
    publication.exclude = [i0 + j + 1000 for j in range(5)]
    publication.exclude_authid = ["user1", "user2", "user3"]
    publication.exclude_authrole = ["roleA", "roleB", "roleC"]

    i0 = util.id()
    publication.eligible = [i0 + j + 1000 for j in range(5)]
    publication.eligible_authid = ["user4", "user5", "user6"]
    publication.eligible_authrole = ["roleD", "roleE", "roleF"]

    publication.enc_algo = Publication.ENC_ALGO_XBR
    publication.enc_key = os.urandom(32)
    publication.enc_serializer = Publication.ENC_SER_CBOR


@pytest.fixture(scope="function")
def publication():
    _publication = Publication()
    fill_publication(_publication)
    return _publication


def test_publication_roundtrip(publication, builder):
    obj = publication.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) in [624, 632, 640]

    _publication = Publication.cast(data)

    assert _publication.timestamp == publication.timestamp
    assert _publication.publication == publication.publication
    assert _publication.publisher == publication.publisher
    assert _publication.topic == publication.topic
    assert _publication.args == publication.args
    assert _publication.kwargs == publication.kwargs
    assert _publication.payload == publication.payload
    assert _publication.acknowledge == publication.acknowledge
    assert _publication.retain == publication.retain
    assert _publication.exclude_me == publication.exclude_me
    assert _publication.exclude == publication.exclude
    assert _publication.exclude_authid == publication.exclude_authid
    assert _publication.exclude_authrole == publication.exclude_authrole
    assert _publication.eligible == publication.eligible
    assert _publication.eligible_authid == publication.eligible_authid
    assert _publication.eligible_authrole == publication.eligible_authrole
    assert _publication.enc_algo == publication.enc_algo
    assert _publication.enc_key == publication.enc_key
    assert _publication.enc_serializer == publication.enc_serializer


def test_publication_roundtrip_perf(publication, builder):
    obj = publication.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"timestamp": 0}

    def loop():
        _publication = Publication.cast(data)
        if True:
            assert _publication.timestamp == publication.timestamp
            assert _publication.publication == publication.publication
            assert _publication.publisher == publication.publisher
            assert _publication.topic == publication.topic
            assert _publication.args == publication.args
            assert _publication.kwargs == publication.kwargs
            assert _publication.payload == publication.payload
            assert _publication.acknowledge == publication.acknowledge
            assert _publication.retain == publication.retain
            assert _publication.exclude_me == publication.exclude_me
            assert _publication.exclude == publication.exclude
            assert _publication.exclude_authid == publication.exclude_authid
            assert _publication.exclude_authrole == publication.exclude_authrole
            assert _publication.eligible == publication.eligible
            assert _publication.eligible_authid == publication.eligible_authid
            assert _publication.eligible_authrole == publication.eligible_authrole
            assert _publication.enc_algo == publication.enc_algo
            assert _publication.enc_key == publication.enc_key
            assert _publication.enc_serializer == publication.enc_serializer

            scratch["timestamp"] += publication.timestamp

    N = 5

    if platform.python_implementation() == "PyPy":
        M = 100000
    else:
        M = 10000
    samples = []
    print("measuring with N={}, M={}:".format(N, M))
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance ({} objects total)".format(ops50, N * M))

    assert ops50 > 1000
    assert scratch["timestamp"] > 0



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrnetwork

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class VerifiedAction(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = VerifiedAction()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsVerifiedAction(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # VerifiedAction
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Globally unique and static ID of action.
    # VerifiedAction
    def Oid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # VerifiedAction
    def OidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # VerifiedAction
    def OidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # VerifiedAction
    def OidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Timestamp (epoch time in ns) of initial creation of this record.
    # VerifiedAction
    def Created(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Verification type.
    # VerifiedAction
    def Vtype(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Verification type.
    # VerifiedAction
    def Vstatus(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Verification code sent.
    # VerifiedAction
    def Vcode(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ID of object of verified action.
    # VerifiedAction
    def VerifiedOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # VerifiedAction
    def VerifiedOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # VerifiedAction
    def VerifiedOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # VerifiedAction
    def VerifiedOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Action data, serialized in CBOR.
    # VerifiedAction
    def VerifiedData(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # VerifiedAction
    def VerifiedDataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # VerifiedAction
    def VerifiedDataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # VerifiedAction
    def VerifiedDataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0


def VerifiedActionStart(builder):
    builder.StartObject(7)


def Start(builder):
    VerifiedActionStart(builder)


def VerifiedActionAddOid(builder, oid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    VerifiedActionAddOid(builder, oid)


def VerifiedActionStartOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOidVector(builder, numElems: int) -> int:
    return VerifiedActionStartOidVector(builder, numElems)


def VerifiedActionAddCreated(builder, created):
    builder.PrependUint64Slot(1, created, 0)


def AddCreated(builder, created):
    VerifiedActionAddCreated(builder, created)


def VerifiedActionAddVtype(builder, vtype):
    builder.PrependUint8Slot(2, vtype, 0)


def AddVtype(builder, vtype):
    VerifiedActionAddVtype(builder, vtype)


def VerifiedActionAddVstatus(builder, vstatus):
    builder.PrependUint8Slot(3, vstatus, 0)


def AddVstatus(builder, vstatus):
    VerifiedActionAddVstatus(builder, vstatus)


def VerifiedActionAddVcode(builder, vcode):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(vcode), 0)


def AddVcode(builder, vcode):
    VerifiedActionAddVcode(builder, vcode)


def VerifiedActionAddVerifiedOid(builder, verifiedOid):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(verifiedOid), 0)


def AddVerifiedOid(builder, verifiedOid):
    VerifiedActionAddVerifiedOid(builder, verifiedOid)


def VerifiedActionStartVerifiedOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartVerifiedOidVector(builder, numElems: int) -> int:
    return VerifiedActionStartVerifiedOidVector(builder, numElems)


def VerifiedActionAddVerifiedData(builder, verifiedData):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(verifiedData), 0)


def AddVerifiedData(builder, verifiedData):
    VerifiedActionAddVerifiedData(builder, verifiedData)


def VerifiedActionStartVerifiedDataVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartVerifiedDataVector(builder, numElems: int) -> int:
    return VerifiedActionStartVerifiedDataVector(builder, numElems)


def VerifiedActionEnd(builder):
    return builder.EndObject()


def End(builder):
    return VerifiedActionEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: user

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# User activation tokens.
class ActivationToken(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ActivationToken()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsActivationToken(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # ActivationToken
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # OID of user object (primary key, never changes).
    # ActivationToken
    def Oid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Activation type.
    # ActivationToken
    def Atype(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Activation status.
    # ActivationToken
    def Status(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Time when the activation was created.
    # ActivationToken
    def Created(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Time when the activation was completed. The status will signal the outcome.
    # ActivationToken
    def Completed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Activation code that was sent.
    # ActivationToken
    def Code(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Email address to which the activation was sent.
    # ActivationToken
    def Email(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # User public key for which this activation was created.
    # ActivationToken
    def Pubkey(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def ActivationTokenStart(builder):
    builder.StartObject(8)


def Start(builder):
    ActivationTokenStart(builder)


def ActivationTokenAddOid(builder, oid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    ActivationTokenAddOid(builder, oid)


def ActivationTokenAddAtype(builder, atype):
    builder.PrependInt8Slot(1, atype, 0)


def AddAtype(builder, atype):
    ActivationTokenAddAtype(builder, atype)


def ActivationTokenAddStatus(builder, status):
    builder.PrependInt8Slot(2, status, 0)


def AddStatus(builder, status):
    ActivationTokenAddStatus(builder, status)


def ActivationTokenAddCreated(builder, created):
    builder.PrependUint64Slot(3, created, 0)


def AddCreated(builder, created):
    ActivationTokenAddCreated(builder, created)


def ActivationTokenAddCompleted(builder, completed):
    builder.PrependUint64Slot(4, completed, 0)


def AddCompleted(builder, completed):
    ActivationTokenAddCompleted(builder, completed)


def ActivationTokenAddCode(builder, code):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(code), 0)


def AddCode(builder, code):
    ActivationTokenAddCode(builder, code)


def ActivationTokenAddEmail(builder, email):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(email), 0)


def AddEmail(builder, email):
    ActivationTokenAddEmail(builder, email)


def ActivationTokenAddPubkey(builder, pubkey):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(pubkey), 0)


def AddPubkey(builder, pubkey):
    ActivationTokenAddPubkey(builder, pubkey)


def ActivationTokenEnd(builder):
    return builder.EndObject()


def End(builder):
    return ActivationTokenEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.gen.meta.DocFormat import DocFormat

from .attribute import Attribute, Attributes
from .schema import Schema

__all__ = ("Schema", "Attribute", "Attributes", "DocFormat")



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrmm

# State of an (off-chain) transaction.
class TransactionState(object):
    # Unset
    NONE = 0
    # The (off-chain) transaction is currently in-flight.
    INFLIGHT = 1
    # The (off-chain) transaction has completed with error (it failed).
    FAILED = 2
    # The (off-chain) transaction has completed with success.
    SUCCESS = 3



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit
import uuid

import txaio

txaio.use_twisted()  # noqa

import numpy as np
import pytest
from txaio import time_ns
from zlmdb import flatbuffers

from cfxdb.tests._util import _gen_ipfs_hash
from cfxdb.xbr import Actor


def fill_actor(actor):
    actor.actor = os.urandom(20)
    actor.actor_type = random.randint(1, 2)
    actor.market = uuid.uuid4()
    actor.timestamp = np.datetime64(time_ns(), "ns")
    actor.joined = random.randint(0, 2**256 - 1)
    actor.security = random.randint(0, 2**256 - 1)
    actor.meta = _gen_ipfs_hash()
    actor.tid = os.urandom(32)
    actor.signature = os.urandom(65)


@pytest.fixture(scope="function")
def actor():
    _actor = Actor()
    fill_actor(_actor)
    return _actor


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def test_actor_roundtrip(actor, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = actor.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 368

    # create python object from bytes (flatbuffes)
    _actor = Actor.cast(data)

    assert _actor.actor == actor.actor
    assert _actor.actor_type == actor.actor_type
    assert _actor.market == actor.market
    assert _actor.timestamp == actor.timestamp
    assert _actor.joined == actor.joined
    assert _actor.security == actor.security
    assert _actor.meta == actor.meta
    assert _actor.tid == actor.tid
    assert _actor.signature == actor.signature


def test_actor_roundtrip_perf(actor, builder):
    obj = actor.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _actor = Actor.cast(data)
        if True:
            assert _actor.actor == actor.actor
            assert _actor.actor_type == actor.actor_type
            assert _actor.market == actor.market
            assert _actor.timestamp == actor.timestamp
            assert _actor.joined == actor.joined
            assert _actor.security == actor.security
            assert _actor.meta == actor.meta
            assert _actor.tid == actor.tid
            assert _actor.signature == actor.signature

            scratch["value"] += _actor.actor_type

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: log

# Managed nodes run-time (live) states.
class MNodeState(object):
    # Unset state.
    NONE = 0
    # Online and fully operational.
    ONLINE = 1
    # Online, but some error occurred and some features or entities might be unavailable.
    DEGRADED = 2



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: arealm

# WAMP URI match type applied (for pattern subscription/registrations and permissions).
class MatchType(object):
    # No match type.
    NONE = 0
    # Match URI exactly.
    EXACT = 1
    # Match by URI prefix.
    PREFIX = 2
    # Match by URI wildcard pattern.
    WILDCARD = 3
